import os
import sys
import inspect
import asyncio
import logging
from pathlib import Path
from promptflow import tool
from types import ModuleType

import semantic_kernel as sk
from semantic_kernel.connectors.ai.open_ai import AzureChatCompletion
from semantic_kernel.planning.sequential_planner import SequentialPlanner
from semantic_kernel.planning.stepwise_planner import StepwisePlanner
from semantic_kernel.planning.action_planner.action_planner import ActionPlanner
from promptflow.connections import AzureOpenAIConnection
from promptflow._core.tool_meta_generator import load_python_module_from_file
from promptflow.contracts.types import FilePath


def get_logger(name: str) -> logging.Logger:
    logger = logging.Logger(name)
    format = logging.Formatter()
    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setFormatter(format)
    logger.addHandler(stdout_handler)
    return logger


def load_classes(code_path: Path):
    module: ModuleType = load_python_module_from_file(code_path)
    target_classes = [m for m in dir(module) 
                      if inspect.isclass(getattr(module, m)) and 
                      inspect.getmodule(getattr(module, m)) == None]
    
    # TODO: assumption of parameterless constructors here....
    # perhaps should use **params as constructor params?
    return [getattr(module, o)() for o in target_classes]


def get_kernel(connection: AzureOpenAIConnection, deployment_name: str, file_path: FilePath):
    code_path = Path(file_path)
    if not code_path.is_absolute():
        code_path = Path(os.getcwd()) / code_path

    if not code_path.exists():
        raise Exception(f"Code file {file_path} does not exist.")
    
    target_classes = load_classes(code_path)
    if len(target_classes) == 0:
        raise Exception(f"No classes found in {code_path}")
    
    # create kernel
    kernel = sk.Kernel(log=get_logger("SEMANTIC_KERNEL"))
    kernel.add_chat_service(
            "chat_completion",
            AzureChatCompletion(
                deployment_name,
                connection.api_base,
                connection.api_key,
            ),
        )
    
    # import function skill definitions
    for c in target_classes:
        kernel.import_skill(c)

    return kernel


@tool
def stepwise(connection: AzureOpenAIConnection, deployment_name: str, file_path: FilePath, intent: str, **params) -> list:
    # get kernel
    kernel = get_kernel(connection, deployment_name, file_path)

    # create planner
    planner = StepwisePlanner(kernel=kernel)

     # create plan
    plan = planner.create_plan(intent)

    # execute plan
    answer = asyncio.run(kernel.run_async(plan)).result

    steps = [{"function": step.name,
              "description": step.description,
              "input": step.parameters.variables,
              "output": step._outputs} for _, step in 
              enumerate(plan._steps)]
    
    return {
        "answer": answer,
        "steps": steps
    }


@tool
def sequential(connection: AzureOpenAIConnection, deployment_name: str, file_path: FilePath, intent: str, **params) -> list:
    # get kernel
    kernel = get_kernel(connection, deployment_name, file_path)

    # create planner
    planner = SequentialPlanner(kernel=kernel)

     # create plan
    plan = asyncio.run(planner.create_plan_async(intent))

    # execute plan
    answer = asyncio.run(kernel.run_async(plan)).result

    steps = [{"function": step.name,
              "description": step.description,
              "input": step.parameters.variables,
              "output": step._outputs} for _, step in 
              enumerate(plan._steps)]
    
    return {
        "answer": answer,
        "steps": steps
    }


@tool
def action(connection: AzureOpenAIConnection, deployment_name: str, file_path: FilePath, intent: str, **params) -> list:
    # get kernel
    kernel = get_kernel(connection, deployment_name, file_path)

    # create planner
    planner = ActionPlanner(kernel=kernel, logger=get_logger("SEMANTIC_KERNEL"))

     # create plan
    plan = asyncio.run(planner.create_plan_async(intent))

    # execute plan
    answer = asyncio.run(kernel.run_async(plan)).result

    return {
        "answer": answer,
        "steps": plan.state
    }



import math
from semantic_kernel.skill_definition import (
    sk_function,
    sk_function_context_parameter,
)
from semantic_kernel.orchestration.sk_context import SKContext


class MathSkills:
    @sk_function(
        description="Takes the square root of a number",
        name="Sqrt",
        input_description="The value to take the square root of",
    )
    def square_root(self, number: str) -> str:
        return str(math.sqrt(float(number)))

    @sk_function(
        description="Adds two numbers together",
        name="Add",
    )
    @sk_function_context_parameter(
        name="input",
        description="The first number to add",
    )
    @sk_function_context_parameter(
        name="input2",
        description="The second number to add",
    )
    def add(self, context: SKContext) -> str:
        return str(float(context["input"]) + float(context["input2"]))

    @sk_function(
        description="Subtract two numbers. If you expect a negative result, the minuend should be the smaller number.",
        name="Subtract",
    )
    @sk_function_context_parameter(
        name="input",
        description="The number to subtract from (i.e., minuend)",
    )
    @sk_function_context_parameter(
        name="subtrahend",
        description="The number to subtract away from the input",
    )
    def subtract(self, context: SKContext) -> str:
        return str(float(context["input"]) - float(context["subtrahend"]))

    @sk_function(
        description="Multiply two numbers together. When increasing by a percentage, don't forget to add 1 to the percentage.",
        name="Multiply",
    )
    @sk_function_context_parameter(
        name="input",
        description="The first number to multiply",
    )
    @sk_function_context_parameter(
        name="input2",
        description="The second number to multiply",
    )
    def multiply(self, context: SKContext) -> str:
        return str(float(context["input"]) * float(context["input2"]))

    @sk_function(
        description="Divide two numbers",
        name="Divide",
    )
    @sk_function_context_parameter(
        name="input",
        description="The first number to divide from (i.e., the numerator)",
    )
    @sk_function_context_parameter(
        name="denominator",
        description="The second number to divide by",
    )
    def divide(self, context: SKContext) -> str:
        return str(float(context["input"]) / float(context["denominator"]))


import os
from pathlib import Path
from skpf.tools.planners import sequential, action, stepwise
from promptflow.connections import AzureOpenAIConnection
from dotenv import load_dotenv
load_dotenv()

API_BASE = os.getenv("API_BASE")
API_KEY = os.getenv("API_KEY")
BASE_PATH = os.path.dirname(os.path.realpath(__file__))


def test_sequential():
    connection = AzureOpenAIConnection(api_key=API_KEY, api_base=API_BASE)
    deployment_name = "gpt-35-turbo"

    code =  (Path(BASE_PATH) / Path("./mathskills.py")).absolute()
    result = sequential(connection, deployment_name, str(code), 
                     "Solve this math problem: what is three plus five divided by eight?")
    assert float(result["answer"]) == 8.0


def test_action():
    connection = AzureOpenAIConnection(api_key=API_KEY, api_base=API_BASE)
    deployment_name = "gpt-35-turbo"

    code =  (Path(BASE_PATH) / Path("./mathskills.py")).absolute()
    result = action(connection, deployment_name, str(code), 
                     "Solve this math problem: what is three plus five divided by eight?")


def test_stepwise():
    connection = AzureOpenAIConnection(api_key=API_KEY, api_base=API_BASE)
    deployment_name = "gpt-35-turbo"

    code =  (Path(BASE_PATH) / Path("./mathskills.py")).absolute()
    result = stepwise(connection, deployment_name, str(code), 
                     "Solve this math problem: what is three plus five?")
    assert "8" in result["answer"]



from setuptools import find_packages, setup

PACKAGE_NAME = "skpf"


setup(
    name="skpf",
    version="0.0.2",
    description="This package contains promptflow tools for working with Smenatic Kernel.",
    packages=find_packages(),
    install_requires=[
        'semantic_kernel'
    ],
    entry_points={
        "package_tools": ["query = skpf.tools.utils:list_package_tools"],
    },
    include_package_data=True,   # This line tells setuptools to include files from MANIFEST.in
)